/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using Zcu.Mve.Visualization;

namespace Zcu.Gsvd.DisplaceMap
{
	/// <summary>
	/// Auxiliary class that provides static methods for finding cell / point attributes.
	/// </summary>
	public class AttributeFinder
	{
		/// <summary>
		/// Triangle mesh cell name.
		/// </summary>
		private static readonly string triName = typeof(Triangle).Name;

		/// <summary>
		/// Searches triangle mesh for cell attributes of given name or type.
		/// </summary>
		/// <remarks>Method first searches for cell attribute of given name.
		/// If no such exists the first attribute of given type is selected.</remarks>
		/// <param name="data">Triangle mesh where to search.</param>
		/// <param name="attrName">Attribute name.</param>
		/// <param name="attrType">Attribute type.</param>
		/// <param name="found">Name of actually selected attribute.</param>
		/// <returns>Returns array of cell attributes. Returns <i>null</i>
		/// if no attribute of given type has been found.</returns>
		public static UniformDataArray FindCellAttrib(
			TriangleMesh data, string attrName, Type attrType, out string found)
		{
			// try to find attribute by name
			if (data.ContainsCellAttrs(triName, attrName))
			{
				// get attribute
				UniformDataArray uda = data.GetCellAttrs(triName, attrName);
				// check attribute type
				if (uda.ElemType == attrType)
				{
					// we found it
					found = attrName;
					return uda;
				}
			}

			// get list of all attributes
			System.Collections.IEnumerator en = data.GetCellAttrsIds(triName).GetEnumerator();
			// go through the list
			while (en.MoveNext())
			{
				// get current attribute
				UniformDataArray uda = data.GetCellAttrs(triName, (string)en.Current);
				// check attribute type
				if (uda.ElemType == attrType)
				{
					// types match
					found = (string)en.Current;
					return uda;
				}
			}

			// no attributes of given type found
			found = "nothing";
			return null;
		}

		/// <summary>
		/// Searches triangle mesh for point attributes of given name or type.
		/// </summary>
		/// <remarks>Method first searches for point attribute of given name.
		/// If no such exists the first attribute of given type is selected.</remarks>
		/// <param name="data">Triangle mesh where to search.</param>
		/// <param name="attrName">Attribute name.</param>
		/// <param name="attrType">Attribute type.</param>
		/// <param name="found">Name of actually selected attribute.</param>
		/// <returns>Returns array of point attributes. Returns <i>null</i>
		/// if no attribute of given type has been found.</returns>
		public static UniformDataArray FindPointAttrib(
			TriangleMesh data, string attrName, Type attrType, out string found)
		{
			// try to find attribute by name
			if (data.ContainsPointAttrs(attrName))
			{
				// get attribute
				UniformDataArray uda = data.GetPointAttrs(attrName);
				// check attribute type
				if (uda.ElemType == attrType)
				{
					// we found it
					found = attrName;
					return uda;
				}
			}

			// get list of all attributes
			System.Collections.IEnumerator en = data.GetPointAttrsIds().GetEnumerator();
			// go through the list
			while (en.MoveNext())
			{
				// get current attribute
				UniformDataArray uda = data.GetPointAttrs((string)en.Current);
				// check attribute type
				if (uda.ElemType == attrType)
				{
					// types match
					found = (string)en.Current;
					return uda;
				}
			}

			// no attributes of given type found
			found = "nothing";
			return null;
		}

		/// <summary>
		/// Searches 2D grid for cell attributes of given name or type.
		/// </summary>
		/// <remarks>Method first searches for cell attribute of given name.
		/// If no such exists the first attribute of given type is selected.</remarks>
		/// <param name="data">2D grid where to search.</param>
		/// <param name="attrName">Attribute name.</param>
		/// <param name="attrType">Attribute type.</param>
		/// <param name="found">Name of actually selected attribute.</param>
		/// <returns>Returns array of cell attributes. Returns <i>null</i>
		/// if no attribute of given type has been found.</returns>
		public static UniformDataArray2D FindCellAttrib(
			Grid2D data, string attrName, Type attrType, out string found)
		{
			UniformDataArray2D uda2D;
			try
			{
				// try to get attribute of given name
				uda2D = data.GetCellAttrs(attrName);
				// check attribute type
				if (uda2D.ElemType == attrType)
				{
					// success
					found = attrName;
					return uda2D;
				}
			}
			catch (Zcu.Mve.Core.MveException)
			{
				// attribute not found, continue searching
			}

			// get list of all attributes
			System.Collections.IEnumerator en = data.GetCellAttrIds().GetEnumerator();
			// go through the list
			while (en.MoveNext())
			{
				// get current attribute
				uda2D = data.GetCellAttrs((string)en.Current);
				// check attribute type
				if (uda2D.ElemType == attrType)
				{
					// types match
					found = (string)en.Current;
					return uda2D;
				}
			}

			// no attributes of given type found
			found = "nothing";
			return null;
		}

		/// <summary>
		/// Searches 2D grid for point attributes of given name or type.
		/// </summary>
		/// <remarks>Method first searches for point attribute of given name.
		/// If no such exists the first attribute of given type is selected.</remarks>
		/// <param name="data">2D grid where to search.</param>
		/// <param name="attrName">Attribute name.</param>
		/// <param name="attrType">Attribute type.</param>
		/// <param name="found">Name of actually selected attribute.</param>
		/// <returns>Returns array of point attributes. Returns <i>null</i>
		/// if no attribute of given type has been found.</returns>
		public static UniformDataArray2D FindPointAttrib(
			Grid2D data, string attrName, Type attrType, out string found)
		{
			UniformDataArray2D uda2D;
			try
			{
				// try to get attribute of given name
				uda2D = data.GetPointAttrs(attrName);
				// check attribute type
				if (uda2D.ElemType == attrType)
				{
					// success
					found = attrName;
					return uda2D;
				}
			}
			catch (Zcu.Mve.Core.MveException)
			{
				// attribute not found, continue searching
			}

			// get list of all attributes
			System.Collections.IEnumerator en = data.GetPointAttrIds().GetEnumerator();
			// go through the list
			while (en.MoveNext())
			{
				// get current attribute
				uda2D = data.GetPointAttrs((string)en.Current);
				// check attribute type
				if (uda2D.ElemType == attrType)
				{
					// types match
					found = (string)en.Current;
					return uda2D;
				}
			}

			// no attributes of given type found
			found = "nothing";
			return null;
		}
	}
}