/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using System.Xml;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Gsvd.DisplaceMap
{
	/// <summary>
	/// <p>Creates a 3D triangle mesh from planar triangle mesh or regular 2D grid.
	/// Elevation (<i>Z</i> coordinate) is computed from <i>XY</i> plane
	/// according to scalar point attributes. Module provides advanced control
	/// over regular 2D grid sizing and positioning. If there is a triangle mesh
	/// at module input the resulting 3D triangle mesh won't have any point or cell
	/// attributes even if input mesh has had some. This is because the third dimension
	/// of created mesh may give attributes different meaning or make them a nonsense.
	/// For example point distance from origin or triangle normal vector would
	/// become incorrect.</p>
	/// <p><b>How it works:</b><br />
	/// In case of planar triangle mesh input things are quite simple.
	/// Points of newly created 3D mesh have the same <i>X</i> and <i>Y</i>
	/// coordinates as those from original planar mesh. <i>Z</i> coordinate
	/// is computed as point elevation (point scalar attribute) multiplied by
	/// elevation factor which may be set in module configuration.</p>
	/// <p>If there is a regular 2D grid at module input some additional computing
	/// is necessary. <i>Z</i> point coordinates are calculated the same way
	/// as in prewious case. Except this we have to determine coordinates
	/// <i>X</i> and <i>Y</i>. The input grid is regular so we can compute them
	/// quite easily according to grid size and position. See module setup
	/// description for further details about possible grid size and position
	/// settings.</p>
	/// <p>Triangle mesh vertices are created in rows from left to right.
	/// Rows are created from bottom to top. Triangles are then made in strips
	/// (not real <i>triangle strips</i>, they are still just isolated triangles)
	/// in the same direction as vertices. Example of point and triangle creation
	/// order is shown in picture below. All triangles have counterclockwise
	/// vertex order.</p>
	/// <img src="TriMeshCreation.png" />
	/// </summary>
	[ModuleInfo("Ji Skla", "Creates a 3D triangle mesh from planar triangle mesh "
		+ "or regular 2D grid. Elevation (Z coordinate) is computed from XY plane "
		+ "according to scalar point attributes. "
		+ "Module provides advanced control over regular 2D grid sizing and positioning.",
		Assemblied = "2005-05-15", Category = "Elevation filter", IconName = "DisplaceMap.ico")]
	[PortInfo("InputMesh", "Regular 2D grid or planar triangle mesh with scalar point attributes.")]
	[PortInfo("OutputMesh", "Elevated 3D triangle mesh.")]
	public class DisplaceMap : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Edge length in X axis direction.
		/// </summary>
		private double sizeX;

		/// <summary>
		/// Edge length in Y axis direction.
		/// </summary>
		private double sizeY;

		/// <summary>
		/// Indicates whether specified size means overall triangle mesh size.
		/// Otherwise size means single grid cell size.
		/// </summary>
		private bool isSizeOverall;

		/// <summary>
		/// Point elevation scale.
		/// </summary>
		private double elevationFactor;

		/// <summary>
		/// Indicates whether regular 2D grid
		/// should be centered in coordinate system origin.
		/// </summary>
		private bool centered;

		/// <summary>
		/// Indicates whether regular 2D grid should have square cells.
		/// </summary>
		private bool square;

		/// <summary>
		/// Triangle mesh that is available at the module output port.
		/// </summary>
		private TriangleMesh outputMesh;

		/// <summary>
		/// Point elevation attribute name.
		/// </summary>
		private string elevAttrName;

		/// <summary>
		/// Determines whether point elevation attribute name is set to default.
		/// </summary>
		private bool elevAttrDefault;

		/// <summary>
		/// Constructor. Creates module input / output ports
		/// and initializes module internal state to default values.
		/// </summary>
		public DisplaceMap()
		{
			// there is no result from previous computation
			outputMesh = null;

			// initialize module default settings
			sizeX = 10.0;
			sizeY = 10.0;
			isSizeOverall = false;
			elevationFactor = 50.0;
			centered = false;
			square = true;

			// initialize elevation attribute name
			elevAttrDefault = true;
			elevAttrName = typeof(Scalar).Name;

			// register module port connection event handler
			OnConnectEvent += new ConnectCallBack(DisplaceMap_OnConnectEvent);

			// add module ports
			AddInPort("InputMesh", typeof(DataSet));
			AddOutPort("OutputMesh", typeof(TriangleMesh));
		}

		/// <summary>
		/// Module port connection event handler. Ensures that only RegGrid2D
		/// or TriangleMesh can be connected at module input port.
		/// </summary>
		/// <param name="source">Module that invoked OnConnectedEvent.</param>
		/// <param name="portFrom">Connection source port.</param>
		/// <param name="portTo">Connection destination port.</param>
		/// <param name="allowed">Determines whether connection is allowed.</param>
		private void DisplaceMap_OnConnectEvent(object source, OutPort portFrom,
			InPort portTo, out bool allowed)
		{
			// accept RegGrid2D or TriangleMesh only
			allowed = portFrom.PortType == typeof(RegGrid2D)
				|| portFrom.PortType == typeof(TriangleMesh);
		}

		/// <summary>
		/// Gets or sets elevation factor.
		/// </summary>
		internal double ElevationFactor
		{
			get { return elevationFactor; }
			set { elevationFactor = value; }
		}

		/// <summary>
		/// Gets point elevation attribute name
		/// and specifies whether default name should be used.
		/// </summary>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		/// <returns>Returns point elevation attribute name.</returns>
		internal string GetElevAttrName(out bool isDefault)
		{
			isDefault = elevAttrDefault;
			return elevAttrName;
		}

		/// <summary>
		/// Sets point elevation attribute name to user or default value.
		/// </summary>
		/// <remarks>If default name should be set,
		/// <i>name</i> parameter is ignored.</remarks>
		/// <param name="name">Point elevation attribute name.</param>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		internal void SetElevAttrName(string name, bool isDefault)
		{
			elevAttrDefault = isDefault;
			if (isDefault)
				// set default value
				elevAttrName = typeof(Scalar).Name;
			else
				// set user value
				elevAttrName = name;
		}

		/// <summary>
		/// Gets or sets whether regular 2D grid should be centered
		/// in coordinate system origin.
		/// </summary>
		internal bool GridCentered
		{
			get { return centered; }
			set { centered = value; }
		}

		/// <summary>
		/// Gets or stes whether regular 2D grid should have square cells.
		/// </summary>
		internal bool SquareCells
		{
			get { return square; }
			set { square = value; }
		}

		/// <summary>
		/// Gets output triangle mesh size setting.
		/// </summary>
		/// <param name="sizeX">Size in X axis direction.</param>
		/// <param name="sizeY">Size in Y axis direction.</param>
		/// <returns>Returns true if size specifies overall mesh size.
		/// Returns false if size specifies single grid cell size.</returns>
		internal bool GetSize(out double sizeX, out double sizeY)
		{
			sizeX = this.sizeX;
			sizeY = this.sizeY;
			return isSizeOverall;
		}

		/// <summary>
		/// Sets output triangle mesh size.
		/// </summary>
		/// <param name="sizeX">Size in X axis direction.</param>
		/// <param name="sizeY">Size in Y axis direction.</param>
		/// <param name="overall">If true, size specifies overall mesh size.
		/// If false, size specifies single grid cell size.</param>
		internal void SetSize(double sizeX, double sizeY, bool overall)
		{
			this.sizeX = sizeX;
			this.sizeY = sizeY;
			this.isSizeOverall = overall;
		}

		/// <summary>
		/// Creates a deep copy of this instance.
		/// </summary>
		/// <returns>Returns a deep copy of this instance.</returns>
		public override Module DeepCopy()
		{
			// create a new instance
			DisplaceMap copy = (DisplaceMap)base.DeepCopy();

			// set to the same state as this
			copy.SetSize(this.sizeX, this.sizeY, this.isSizeOverall);
			copy.elevationFactor = this.elevationFactor;
			copy.centered = this.centered;
			copy.square = this.square;
			copy.SetElevAttrName(this.elevAttrName, this.elevAttrDefault);

			return copy;
		}

		/// <summary>
		/// Saves current module settings.
		/// </summary>
		/// <param name="config">XML element where to save.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			// save grid size
			XmlElement gridSize = config.OwnerDocument.CreateElement("gridSize");
			gridSize.SetAttribute("x", sizeX.ToString());
			gridSize.SetAttribute("y", sizeY.ToString());
			gridSize.SetAttribute("overall", isSizeOverall.ToString());
			config.AppendChild(gridSize);

			// save elevation factor
			XmlElement elevFactor = config.OwnerDocument.CreateElement("elevationFactor");
			elevFactor.InnerText = elevationFactor.ToString();
			config.AppendChild(elevFactor);

			// save centered flag
			XmlElement cntrd = config.OwnerDocument.CreateElement("centered");
			cntrd.InnerText = centered.ToString();
			config.AppendChild(cntrd);

			// save square flag
			XmlElement sqr = config.OwnerDocument.CreateElement("square");
			sqr.InnerText = square.ToString();
			config.AppendChild(sqr);

			// save elevation attribute
			XmlElement elevAttr = config.OwnerDocument.CreateElement("elevationAttribute");
			elevAttr.SetAttribute("default", elevAttrDefault.ToString());
			elevAttr.InnerText = elevAttrName;
			config.AppendChild(elevAttr);
		}

		/// <summary>
		/// Loads module settings.
		/// </summary>
		/// <param name="config">XML element from where to load.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			// load module setup
			foreach (XmlElement element in config.ChildNodes)
				switch (element.Name)
				{
					case "gridSize":
						// grid size definition
						string attribGrid = element.GetAttribute("x");
						sizeX = Parser.DoubleValue(attribGrid, this.Name);
						attribGrid = element.GetAttribute("y");
						sizeY = Parser.DoubleValue(attribGrid, this.Name);
						attribGrid = element.GetAttribute("overall");
						isSizeOverall = Parser.BoolValue(attribGrid, this.Name);
						break;

					case "elevationFactor":
						// elevation factor definition
						elevationFactor = Parser.DoubleValue(element.InnerText, this.Name);
						break;

					case "centered":
						// load centered flag
						centered = Parser.BoolValue(element.InnerText, this.Name);
						break;

					case "square":
						// load square flag
						square = Parser.BoolValue(element.InnerText, this.Name);
						break;

					case "elevationAttribute":
						// get "default" attribute
						string attrib = element.GetAttribute("default");
						// try to parse attribute value
						elevAttrDefault = Parser.BoolValue(attrib, this.Name);
						// get elevation attribute name
						elevAttrName = element.InnerText.Trim();
						break;

					default:
						// unknown config tag
						Console.WriteLine(this.Name + " warning:");
						Console.WriteLine("\"" + element.Name + "\" is not recognized "
							+ "as known module configuration tag - ignoring.");
						break;
				}	// end of switch (element.Name)
			// end of foreach
		}

		/// <summary>
		/// <p><i>DisplaceMap</i> module setup dialog.</p>
		/// <img src="DisplaceMapSetup.png" />
		/// <p><i>Elevation factor</i> sets the multiplier of point elevation
		/// attributes. It's in fact point elevation scale. <i>Elevation attribute name</i>
		/// is the name for scalar point elevation attributes. It is recomended
		/// to use the default value unless you really need different attribute name.</p>
		/// <p>The panel below provides settings for regular 2D grid input.
		/// <i>Output mesh position</i> determines where in the <i>XY</i> plane
		/// will be the output triangle mesh placed. It may be centered in the origin
		/// or start in the origin and continue to the first quadrant. See example
		/// in the image below.</p>
		/// <img src="RegGridPosition.png" />
		/// <p>The grid size may be set in two ways. First is to enter dimensions
		/// of a single grid cell. If you want cells to be square just check appropriate
		/// check box and enter <i>X</i> edge size only. Another way is to specify
		/// overall grid dimensions. Again you may only enter <i>X</i> edge size
		/// and check the check box to maintain original grid aspect ratio.</p>
		/// </summary>
		/// <remarks>Creates module setup dialog.</remarks>
		/// <returns>Returns module setup dialog reference.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new DisplaceMapSetup(this);
		}

		/// <summary>
		/// Creates 3D triangle mesh from input planar triangle mesh or regular 2D grid.
		/// </summary>
		public override void Execute()
		{
			// check whether there is some new input
			if (IsDataSame("InputMesh") && outputMesh != null)
			{
				// publish old output
				SetOutput("OutputMesh", outputMesh, true);
				// no more work
				return;
			}

			// get input data
			DataSet input = (DataSet)GetInput("InputMesh");

			// process input regular 2D grid or triangle mesh
			if (input is RegGrid2D)
				outputMesh = ProcessInput((RegGrid2D)input);
			else
				outputMesh = ProcessInput((TriangleMesh)input);

			// publish new output
			SetOutput("OutputMesh", outputMesh, false);
		}

		/// <summary>
		/// Creates 3D triangle mesh from regular 2D grid. Elevation (Z coordinate)
		/// is computed from XY plane according to scalar point attributes.
		/// </summary>
		/// <param name="inputGrid">Input regular 2D grid.</param>
		/// <returns>Returns elevated triangle mesh.</returns>
		/// <remarks>Created triangles have counterclockwise point order.</remarks>
		private TriangleMesh ProcessInput(RegGrid2D inputGrid)
		{
			string attribFound;
			// get point elevation attributes
			UniformDataArray2D elevation = AttributeFinder.FindPointAttrib(
				inputGrid, elevAttrName, typeof(Scalar), out attribFound);

			// check success
			if (elevation == null)
				// there are no point elevation attributes in input grid
				throw new MveException(this.Name + ":\n"
					+ "Input regular 2D grid does not contain any scalar point attributes.");

			// check we found what we wanted
			if (attribFound != elevAttrName)
			{
				// show a warning that given attribute has not been found
				Console.WriteLine(this.Name + " warning:");
				Console.WriteLine("\"" + elevAttrName + "\" point attribute not found or has wrong type.");
				Console.WriteLine("Using \"" + attribFound + "\" attribute instead.");
			}

			// create regular triangle mesh points
			UniformDataArray triMeshPoints
				= CreateRegTriMeshPoints(elevation, inputGrid.Width + 1, inputGrid.Height + 1);

			// create regular triangle mesh cells
			UniformDataArray triMeshCells
				= CreateRegTriMeshCells(inputGrid.Width + 1, inputGrid.Height + 1);

			// prepare data structure for output
			TriangleMesh output = new TriangleMesh();

			// store output points
			output.Points = triMeshPoints;
			// store cell information
			output.AddCells(typeof(Triangle).Name, triMeshCells);

			return output;
		}

		/// <summary>
		/// Creates regular triangle mesh points.
		/// </summary>
		/// <param name="elevation">Point elevations.</param>
		/// <param name="width">Triangle mesh width (number of points).</param>
		/// <param name="height">Triangle mesh height (number of points).</param>
		/// <returns>Returns regular triangle mesh points.</returns>
		private UniformDataArray CreateRegTriMeshPoints(UniformDataArray2D elevation,
			int width, int height)
		{
			// prepare memory for output points
			UniformDataArray meshPoints
				= new UniformDataArray(typeof(Point3D), width * height);

			#region Miscellaneous auxiliary computing

			// initialize single cell size
			double gridCellSizeX = sizeX;
			double gridCellSizeY = sizeY;
			// check size semantics
			if (isSizeOverall)
			{
				// compute single cell size
				gridCellSizeX /= (double)(width - 1);
				gridCellSizeY /= (double)(height - 1);
			}
			// check whether cells should be square
			if (square)
				// set cell size to square
				gridCellSizeY = gridCellSizeX;

			// coordinate offset
			double centerX = 0;
			double centerY = 0;
			// check grid positioning
			if (centered)
			{
				// center result if desired
				centerX = - (width - 1) * gridCellSizeX / 2.0;
				centerY = - (height - 1) * gridCellSizeY / 2.0;
			}

			#endregion

			// we could compute array indices inside for loops,
			// but this will be easier and faster
			int pointIndex = 0;

			// create output triangle mesh points
			for (int j = 0; j < height; j++)
				for (int i = 0; i < width; i++)
				{
					// compute point coordinates
					double x = i * gridCellSizeX + centerX;
					double y = j * gridCellSizeY + centerY;
					double z = (Scalar)elevation[height-1-j, i] * elevationFactor;
					// height-1-j --> we must flip picture Y coordinate

					// store point
					meshPoints[pointIndex] = new Point3D(x, y, z);
					pointIndex++;
				}

			return meshPoints;
		}

		/// <summary>
		/// Creates regular triangle mesh cells.
		/// </summary>
		/// <remarks>Triangles are created with counterclockwise point order.</remarks>
		/// <param name="width">Triangle mesh width (number of points).</param>
		/// <param name="height">Triangle mesh height (number of points).</param>
		/// <returns>Returns regular triangle mesh cells.</returns>
		private UniformDataArray CreateRegTriMeshCells(int width, int height)
		{
			// prepare memory for cells
			UniformDataArray meshCells
				= new UniformDataArray(typeof(Triangle), 2 * (width-1) * (height-1));

			// we could compute array indices inside for loops,
			// but this will be easier and faster
			int triangleIndex = 0;

			// create appropriate cells
			// triangles are created with counterclockwise point order
			for (int j = 0; j < height-1; j++)
				for (int i = 0; i < width-1; i++)
				{
					// compute base point index
					int baseIndex = i + j * width;

					// create lower left triangle
					meshCells[triangleIndex]
						= new Triangle(baseIndex, baseIndex + 1, baseIndex + width);
					// create upper right triangle
					meshCells[triangleIndex + 1]
						= new Triangle(baseIndex + width, baseIndex + 1, baseIndex + width + 1);

					// increment triangle counter, we added 2 triangles
					triangleIndex += 2;
				}

			return meshCells;
		}

		/// <summary>
		/// Creates 3D triangle mesh from a planar triangle mesh. Elevation (Z coordinate)
		/// is computed from XY plane according to scalar point attributes.
		/// </summary>
		/// <param name="inputMesh">Input triangle mesh.</param>
		/// <returns>Returns elevated triangle mesh.</returns>
		private TriangleMesh ProcessInput(TriangleMesh inputMesh)
		{
			// get input triangle mesh points
			UniformDataArray inputPoints = inputMesh.Points;

			// check input data format
			if (inputPoints.ElemType != typeof(Point2D))
				throw new MveException(this.Name + ":\n"
					+ "Input triangle mesh must have points typed to Point2D.");

			string attribFound;
			// get point elevation attributes
			UniformDataArray elevation = AttributeFinder.FindPointAttrib(
				inputMesh, elevAttrName, typeof(Scalar), out attribFound);

			// check success
			if (elevation == null)
				// there are no point elevation attributes in input triangle mesh
				throw new MveException(this.Name + ":\n"
					+ "Input triangle mesh does not contain any scalar point attributes.");

			// check we found what we wanted
			if (attribFound != elevAttrName)
			{
				// show a warning that given attribute has not been found
				Console.WriteLine(this.Name + " warning:");
				Console.WriteLine("\"" + elevAttrName + "\" point attribute not found or has wrong type.");
				Console.WriteLine("Using \"" + attribFound + "\" attribute instead.");
			}

			// prepare memory for output points
			UniformDataArray outputPoints
				= new UniformDataArray(typeof(Point3D), inputPoints.Count);

			// process all points
			for (int i = 0; i < inputPoints.Count; i++)
			{
				// get current input point
				Point2D current = (Point2D)inputPoints[i];
				// compute output point height
				double z = (Scalar)elevation[i] * elevationFactor;
				// store new output point
				outputPoints[i] = new Point3D(current.X, current.Y, z);
			}

			// prepare data structure for output
			TriangleMesh output = new TriangleMesh();

			// store output points
			output.Points = outputPoints;
			// copy cells from input (shallow copy)
			output.AddCells(typeof(Triangle).Name, inputMesh.GetCells(typeof(Triangle).Name));

			return output;
		}
	}
}