/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using System.Xml;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Gsvd.DisplaceMap
{
	/// <summary>
	/// Available color mapping functions.
	/// </summary>
	internal enum ColorMapping {Classic, Burn, Gray, Xray, Copper, Sepia};

	/// <summary>
	/// <p>Paints triangle mesh points according to their elevation.
	/// Direction of elevation may be specified arbitrary in module setup dialog
	/// or by Vector3D connected to module input.</p>
	/// <p>Several color mappings are available. If input triangle mesh is planar
	/// and perpendicular to the direction of elevation then all points are painted
	/// in the same color that corresponds to the lowest elevation level.</p>
	/// <p><b>How it works:</b><br />
	/// Algorithm proceeds it two passes. In the first pass an elevation level
	/// for each point is computed. The direction of elevation yields
	/// a normal vector that designates parallel planes. Each point then lies
	/// within one of these planes and it determines it's elevation level.</p>
	/// <p>In the second pass color attributes are assigned to points according
	/// to their elevation. Point colors are determined by selected
	/// color mapping scheme which is adapted to fit point elevation range.</p>
	/// </summary>
	[ModuleInfo("Ji Skla", "Paints triangle mesh points according to their elevation. "
		+ "Direction of elevation may be specified arbitrary in module setup dialog "
		+ "or by Vector3D connected to module input. "
		+ "Several color mappings are available.",
		IconName = "Elevation.ico", Assemblied = "2005-05-16", Category = "Color Filter")]
	[PortInfo("InputMesh", "Triangle mesh to paint.")]
	[PortInfo("ElevationVector", "Optional. Direction in which the elevation will be computed.")]
	[PortInfo("OutputMesh", "Triangle mesh with point color attributes.")]
	public class Elevation : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Color mapping function delegate declaration.
		/// </summary>
		private delegate ColorRGBA ColorMappingFunction(float elevation);

		/// <summary>
		/// Color mapping function delegate instance.
		/// </summary>
		private ColorMappingFunction ColorMap;

		/// <summary>
		/// Color mapping function name.
		/// </summary>
		private ColorMapping colorMapping;

		/// <summary>
		/// Triangle mesh that is available at the module output port.
		/// </summary>
		private TriangleMesh outputMesh;

		/// <summary>
		/// Vector that defines direction in which the elevation will be computed.
		/// </summary>
		private Vector3D elevationVector;

		/// <summary>
		/// User defined direction in which the elevation will be computed.
		/// </summary>
		/// <remarks>
		/// This field is necessary to back up user defined vector
		/// when using the vector from module input port.
		/// </remarks>
		private Vector3D userElevVector;

		/// <summary>
		/// Triangle mesh cell color attribute name.
		/// </summary>
		private string colorAttrName;

		/// <summary>
		/// Determines whether triangle mesh
		/// cell color attribute name is set to default.
		/// </summary>
		private bool colorAttrDefault;

		/// <summary>
		/// Constructor. Creates module input / output ports
		/// and initializes module internal state.
		/// </summary>
		public Elevation()
		{
			// initialize elevation vector
			userElevVector = new Vector3D(0, 0, 1);

			// initialize color mapping function
			ColorMap = new ColorMappingFunction(ColorMapClassic);
			colorMapping = ColorMapping.Classic;

			// initialize attribute name
			colorAttrDefault = true;
			colorAttrName = typeof(ColorRGBA).Name;

			// there is no result from previous computation
			outputMesh = null;

			// add module ports
			AddInPort("InputMesh", typeof(TriangleMesh));
			AddInPort("ElevationVector", typeof(Vector3D), false);
			AddOutPort("OutputMesh", typeof(TriangleMesh));
		}

		/// <summary>
		/// Gets or sets the elevation vector.
		/// </summary>
		internal Vector3D UserElevationVector
		{
			get { return (Vector3D)userElevVector.DeepCopy(); }
			set { userElevVector = (Vector3D)value.DeepCopy(); }
		}

		/// <summary>
		/// Gets or sets color mapping function.
		/// </summary>
		internal ColorMapping SelectedColorMapping
		{
			get { return colorMapping; }

			set
			{
				// store color mapping name
				colorMapping = value;
				// determine and chose appropriate function
				switch (value)
				{
					case ColorMapping.Classic:
						ColorMap = new ColorMappingFunction(ColorMapClassic);
						break;

					case ColorMapping.Burn:
						ColorMap = new ColorMappingFunction(ColorMapBurn);
						break;

					case ColorMapping.Gray:
						ColorMap = new ColorMappingFunction(ColorMapGray);
						break;

					case ColorMapping.Xray:
						ColorMap = new ColorMappingFunction(ColorMapXray);
						break;

					case ColorMapping.Copper:
						ColorMap = new ColorMappingFunction(ColorMapCopper);
						break;

					case ColorMapping.Sepia:
						ColorMap = new ColorMappingFunction(ColorMapSepia);
						break;
				}
			}
		}

		/// <summary>
		/// Gets color attribute name and specifies whether default name should be used.
		/// </summary>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		/// <returns>Returns color attribute name.</returns>
		internal string GetColorAttrName(out bool isDefault)
		{
			isDefault = colorAttrDefault;
			return colorAttrName;
		}

		/// <summary>
		/// Sets color attribute name to user or default value.
		/// </summary>
		/// <remarks>If default name should be set,
		/// <i>name</i> parameter is ignored.</remarks>
		/// <param name="name">Color attribute name.</param>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		internal void SetColorAttrName(string name, bool isDefault)
		{
			colorAttrDefault = isDefault;
			if (isDefault)
				// set default value
				colorAttrName = typeof(ColorRGBA).Name;
			else
				// set user value
				colorAttrName = name;
		}

		/// <summary>
		/// Creates a deep copy of this instance.
		/// </summary>
		/// <returns>Returns a deep copy of this instance.</returns>
		public override Module DeepCopy()
		{
			// create a new instance
			Elevation copy = (Elevation)base.DeepCopy();

			// set to the same state as this
			copy.SelectedColorMapping = this.colorMapping;	// also sets function delegate
			copy.UserElevationVector = this.userElevVector;
			copy.SetColorAttrName(this.colorAttrName, this.colorAttrDefault);

			return copy;
		}

		/// <summary>
		/// Saves current module settings.
		/// </summary>
		/// <param name="config">XML element where to save.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			// save user elevation vector
			XmlElement vector = config.OwnerDocument.CreateElement("elevationVector");
			vector.SetAttribute("x", userElevVector.X.ToString());
			vector.SetAttribute("y", userElevVector.Y.ToString());
			vector.SetAttribute("z", userElevVector.Z.ToString());
			config.AppendChild(vector);

			// save selected color mapping
			XmlElement colorMap = config.OwnerDocument.CreateElement("colorMapping");
			switch (colorMapping)
			{
				case ColorMapping.Classic:
					colorMap.InnerText = "Classic";
					break;

				case ColorMapping.Burn:
					colorMap.InnerText = "Burn";
					break;

				case ColorMapping.Gray:
					colorMap.InnerText = "Gray";
					break;

				case ColorMapping.Xray:
					colorMap.InnerText = "Xray";
					break;

				case ColorMapping.Copper:
					colorMap.InnerText = "Copper";
					break;

				case ColorMapping.Sepia:
					colorMap.InnerText = "Sepia";
					break;
			}
			config.AppendChild(colorMap);

			// save color attribute name setting
			XmlElement colorAttr = config.OwnerDocument.CreateElement("colorAttribute");
			colorAttr.SetAttribute("default", colorAttrDefault.ToString());
			colorAttr.InnerText = colorAttrName;
			config.AppendChild(colorAttr);
		}

		/// <summary>
		/// Loads module settings.
		/// </summary>
		/// <param name="config">XML element from where to load.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			// load module setup
			foreach (XmlElement element in config.ChildNodes)
				switch (element.Name)
				{
					case "elevationVector":
						// get and parse elevation vector components
						string coord = element.GetAttribute("x");
						double x = Parser.DoubleValue(coord, this.Name);
						coord = element.GetAttribute("y");
						double y = Parser.DoubleValue(coord, this.Name);
						coord = element.GetAttribute("z");
						double z = Parser.DoubleValue(coord, this.Name);
						// create vector
						userElevVector = new Vector3D(x, y, z);
						break;

					case "colorMapping":
						// color mapping definition
						switch (element.InnerText.Trim().ToLower())
						{
							case "classic":
								ColorMap = new ColorMappingFunction(ColorMapClassic);
								colorMapping = ColorMapping.Classic;
								break;

							case "burn":
								ColorMap = new ColorMappingFunction(ColorMapBurn);
								colorMapping = ColorMapping.Burn;
								break;

							case "gray":
								ColorMap = new ColorMappingFunction(ColorMapGray);
								colorMapping = ColorMapping.Gray;
								break;

							case "xray":
								ColorMap = new ColorMappingFunction(ColorMapXray);
								colorMapping = ColorMapping.Xray;
								break;

							case "copper":
								ColorMap = new ColorMappingFunction(ColorMapCopper);
								colorMapping = ColorMapping.Copper;
								break;

							case "sepia":
								ColorMap = new ColorMappingFunction(ColorMapSepia);
								colorMapping = ColorMapping.Sepia;
								break;

							default:
								// unknown function keyword
								throw new MveException(this.Name + ":\n"
									+ "Error reading module configuration. \""
									+ element.InnerText.Trim()
									+ "\" is not recognized as known color mapping function.");
								// break;	// unreachable code
						}
						break;

					case "colorAttribute":
						// get "default" attribute
						string attrib = element.GetAttribute("default");
						// try to parse attribute value
						colorAttrDefault = Parser.BoolValue(attrib, this.Name);
						// get color attribute name
						colorAttrName = element.InnerText.Trim();
						break;

					default:
						// unknown config tag
						Console.WriteLine(this.Name + " warning:");
						Console.WriteLine("\"" + element.Name + "\" is not recognized "
							+ "as known module configuration tag - ignoring.");
						break;
				}	// end of switch (element.Name)
			// end of foreach
		}

		/// <summary>
		/// <p><i>Elevation</i> module setup dialog.</p>
		/// <img src="ElevationSetup.png" />
		/// <p><i>Elevation vector</i> specifies direction in which point elevation
		/// will be computed. The three buttons below helps quickly align
		/// vectors with one of coordinate system axes. Elevation vector setting
		/// is ignored if there is a vector connected to module input port.</p>
		/// <p><i>Color attribute name</i> is the name for color attributes
		/// that will be assigned to points. It is recomended to use the default value
		/// unless you really need different attribute name.</p>
		/// <p>In the <i>Color mapping</i> panel you can choose which
		/// color mapping scheme will be applied to triangle mesh points.
		/// If you want inverse mapping just reverse the elevation vector.</p>
		/// </summary>
		/// <remarks>Creates module setup dialog.</remarks>
		/// <returns>Returns module setup dialog reference.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new ElevationSetup(this);
		}

		/// <summary>
		/// Paints triangle mesh points according to their elevation.
		/// </summary>
		public override void Execute()
		{
			// check whether there is some new input
			if (IsDataSame("InputMesh") && outputMesh != null)
			{
				// publish old output
				SetOutput("OutputMesh", outputMesh, true);
				// no more work
				return;
			}

			// check for elevation vector at input port
			if (GetInput("ElevationVector") != null)
				// use vector from module input port
				elevationVector = (Vector3D)GetInput("ElevationVector");
			else
				// use user defined vector
				elevationVector = userElevVector;

			// get input data
			TriangleMesh inputMesh = (TriangleMesh)GetInput("InputMesh");

			// make a shallow copy
			outputMesh = new TriangleMesh();
			// copy library names
			foreach (string libName in inputMesh.LibNames)
				outputMesh.AddLibName(libName);
			// copy point array reference
			outputMesh.Points = inputMesh.Points;
			// copy point attribute references
			foreach (string attribID in inputMesh.GetPointAttrsIds())
				outputMesh.AddPointAttrs(attribID, inputMesh.GetPointAttrs(attribID));
			// copy cell array reference
			string triName = typeof(Triangle).Name;
			outputMesh.AddCells(triName, inputMesh.GetCells(triName));
			// copy cell attribute references
			foreach (string attribID in inputMesh.GetCellAttrsIds(triName))
				outputMesh.AddCellAttrs(triName, attribID, inputMesh.GetCellAttrs(triName, attribID));

			// compute point colors
			UniformDataArray pointColorAttrs = ComputePointColors(inputMesh);

			// assign attributes to points
			outputMesh.AddPointAttrs(colorAttrName, pointColorAttrs);

			// publish new output
			SetOutput("OutputMesh", outputMesh, false);
		}

		/// <summary>
		/// Computes triangle mesh point color attributes according to point elevation.
		/// </summary>
		/// <param name="inputTriMesh">Input triangle mesh.</param>
		/// <returns>Returns array of point color attributes.</returns>
		private UniformDataArray ComputePointColors(TriangleMesh inputTriMesh)
		{
			// get the number of points
			int pointCount = inputTriMesh.Points.Count;

			// prepare array for point elevation
			double[] elevation = new double[pointCount];
			// initialize minimum and maximum
			double min = double.MaxValue;
			double max = double.MinValue;

			// prepare array for point color attributes
			UniformDataArray pointColors
				= new UniformDataArray(typeof(ColorRGBA), pointCount);

			// Pass 1 - Find point elevation range
			for (int i = 0; i < pointCount; i++)
			{
				// get current point
				Point3D current = (Point3D)inputTriMesh.Points[i];

				// compute elevation
				double elev = elevation[i] = ComputeElevation(current);

				// check elevation range
				if (elev < min)
					min = elev;
				if (elev > max)
					max = elev;
			}

			// compute elevation interval range
			double intervalRange = max - min;
			// check that max > min, we don't want to divide by zero
			if (intervalRange == 0)
				intervalRange = 1;

			// Pass 2 - Set point colors with respect to their elevation
			for (int i = 0; i < pointCount; i++)
			{
				// normalize point elevation to [0; 1] interval
				double elev = (elevation[i] - min) / intervalRange;
				// compute point color
				pointColors[i] = ColorMap((float)elev);
			}

			return pointColors;
		}

		/// <summary>
		/// Computes point elevation in selected direction.
		/// </summary>
		/// <param name="point">Point to compute elevation of.</param>
		/// <returns>Returns point elevation.</returns>
		private double ComputeElevation(Point3D point)
		{
			return elevationVector.X * point.X
				+ elevationVector.Y * point.Y
				+ elevationVector.Z * point.Z;
		}

		#region Color mapping functions

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapClassic(float elevation)
		{
			float r, g, b;

			// compute color channels
			if (elevation < 0.5)
			{
				r = 0;
				g = elevation * 2;
				b = (0.5f - elevation) * 2;
			}
			else
			{
				r = (elevation - 0.5f) * 2;
				g = (1 - elevation) * 2;
				b = 0;
			}

			return new ColorRGBA(r, g, b);
		}

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapBurn(float elevation)
		{
			float r, g, b;

			const float THIRD = 1f / 3f;
			const float TWO_THIRDS = 2f / 3f;

			// compute color channels
			if (elevation < THIRD)
			{
				r = elevation * 3;
				g = 0;
				b = 0;
			}
			else if (elevation < TWO_THIRDS)
			{
				r = 1;
				g = (elevation - THIRD) * 3;
				b = 0;
			}
			else
			{
				r = 1;
				g = 1;
				b = (elevation - TWO_THIRDS) * 3;
			}

			return new ColorRGBA(r, g, b);
		}

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapGray(float elevation)
		{
			return new ColorRGBA(elevation, elevation, elevation);
		}

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapXray(float elevation)
		{
			float r, g, b;

			const float THIRD = 1f / 3f;
			const float TWO_THIRDS = 2f / 3f;

			// compute color channels
			// same scheme as for burn color mapping
			if (elevation < THIRD)
			{
				r = elevation * 3;
				g = 0;
				b = 0;
			}
			else if (elevation < TWO_THIRDS)
			{
				r = 1;
				g = (elevation - THIRD) * 3;
				b = 0;
			}
			else
			{
				r = 1;
				g = 1;
				b = (elevation - TWO_THIRDS) * 3;
			}

			// swap red and blue, add 7x gray color mapping, compute average
			r = (b + 7f*elevation) / 8f;
			g = (g + 7f*elevation) / 8f;
			b = (r + 7f*elevation) / 8f;

			return new ColorRGBA(r, g, b);
		}

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapCopper(float elevation)
		{
			float r, g, b;

			// compute color channels
			if (elevation < 0.8)
				r = elevation * 1.25f;
			else
				r = 1;

			g = elevation * 0.8f;
			b = elevation * 0.5f;

			return new ColorRGBA(r, g, b);
		}

		/// <summary>
		/// Computes point color based on it's elevation.
		/// </summary>
		/// <param name="elevation">Point elevation normalized into [0; 1] interval.</param>
		/// <returns>Returns point color.</returns>
		private ColorRGBA ColorMapSepia(float elevation)
		{
			float r, g, b;

			const float THIRD = 1f / 3f;
			const float TWO_THIRDS = 2f / 3f;

			// compute color channels
			// same scheme as for burn color mapping
			if (elevation < THIRD)
			{
				r = elevation * 3;
				g = 0;
				b = 0;
			}
			else if (elevation < TWO_THIRDS)
			{
				r = 1;
				g = (elevation - THIRD) * 3;
				b = 0;
			}
			else
			{
				r = 1;
				g = 1;
				b = (elevation - TWO_THIRDS) * 3;
			}

			// add 2x gray color mapping, compute average, compute square root
			r = (float)Math.Sqrt((r + 2f*elevation) / 3f);
			g = (float)Math.Sqrt((g + 2f*elevation) / 3f);
			b = (float)Math.Sqrt((b + 2f*elevation) / 3f);

			return new ColorRGBA(r, g, b);
		}

		#endregion
	}
}