/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using MyVertex = Microsoft.DirectX.Direct3D.CustomVertex.PositionColored;

namespace Zcu.Gsvd.DisplaceMap.Testing
{
	/// <summary>
	/// Renderer window with mouse controllable view.
	/// </summary>
	public class RendererWindow : System.Windows.Forms.Form
	{
		/// <summary>
		/// Direct3D device.
		/// </summary>
		private Device device = null;

		/// <summary>
		/// Triangle mesh vertices.
		/// </summary>
		private MyVertex[] vertices;

		/// <summary>
		/// Triangle mesh vertex indices.
		/// </summary>
		private int[] indices;

		/// <summary>
		/// Matrix describing view rotation.
		/// </summary>
		private Matrix rotationMatrix;

		/// <summary>
		/// Camera distance from origin.
		/// </summary>
		private float distance;

		/// <summary>
		/// Mouse cursor X coordinate memory.
		/// </summary>
		private int oldMouseX;

		/// <summary>
		/// Mouse cursor Y coordinate memory.
		/// </summary>
		private int oldMouseY;

		/// <summary>
		/// Determines whether triangle mesh should be rendered as wireframe.
		/// </summary>
		private bool wireframe;

		/// <summary>
		/// Constructor. Initializes render window and stores triangle mesh.
		/// </summary>
		/// <param name="vertices">Triangle mesh vertices.</param>
		/// <param name="indices">Triangle mesh vertex indices.</param>
		public RendererWindow(MyVertex[] vertices, int[] indices)
		{
			// set window properties
			this.Size = new System.Drawing.Size(300,300);
			this.Text = "Triangle mesh renderer";

			// initialize
			oldMouseX = 0;
			oldMouseY = 0;
			wireframe = false;

			// register window event handlers
			this.MouseMove += new MouseEventHandler(RendererWindow_MouseMove);
			this.KeyDown += new KeyEventHandler(RendererWindow_KeyDown);

			// store triangle mesh
			this.vertices = vertices;
			this.indices = indices;

			// initialize D3D
			InitDirect3D();
		}

		/// <summary>
		/// Standard rendering loop.
		/// </summary>
		public void RenderingLoop()
		{
			while (this.Created)
			{
				// respond to events
				Application.DoEvents();
				// refresh scene
				Render();
			}
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				// release D3D device
				if (device != null)
				{
					device.Dispose();
					device = null;
				}
			}

			base.Dispose(disposing);
		}

		/// <summary>
		/// Paints window background.
		/// </summary>
		protected override void OnPaintBackground(PaintEventArgs pevent)
		{
			// do not paint background
		}

		/// <summary>
		/// Paints window content.
		/// </summary>
		protected override void OnPaint(PaintEventArgs e)
		{
			Render();
		}

		/// <summary>
		/// Initializes Direct3D device and prepares window for rendering.
		/// </summary>
		private void InitDirect3D()
		{
			// initialize camera properties
			distance = - 2.5f;
			rotationMatrix = Matrix.Identity;

			// prepare D3D device parameters
			PresentParameters presParams = new PresentParameters();
			// no fullscreen
			presParams.Windowed = true;
			// discard old buffer content on swap
			presParams.SwapEffect = SwapEffect.Discard;
			// enable depth stencil
			presParams.EnableAutoDepthStencil = true;
			// set 16bit depth stencil
			presParams.AutoDepthStencilFormat = DepthFormat.D16;

			// create a hardware device, render to this window
			device = new Device(0, DeviceType.Hardware, this,
				CreateFlags.MixedVertexProcessing, presParams);
			// register device reset event handler
			device.DeviceReset += new EventHandler(OnDeviceReset);
			// make device ready
			OnDeviceReset(device, null);
		}

		/// <summary>
		/// D3D device reset event handler. Restores D3D device settings.
		/// </summary>
		private void OnDeviceReset(object sender, EventArgs e)
		{
			Device dev = (Device)sender;

			// Set rendering properties
			// disable lighting
			dev.RenderState.Lighting = false;
			// disable backface culling
			dev.RenderState.CullMode = Cull.None;
			// enable depth buffering
			dev.RenderState.ZBufferEnable = true;
			// set up a perspective projection
			dev.Transform.Projection
				= Matrix.PerspectiveFovRH((float)Math.PI / 4f, 1, 0.05f, 10);
			// do not transform world
			dev.Transform.World = Matrix.Identity;
		}

		/// <summary>
		/// Renders the triangle mesh.
		/// </summary>
		private void Render()
		{
			if (device == null)
				return;

			// clear scene
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer,
				System.Drawing.Color.DarkBlue.ToArgb(), 1, 0);
			// start rendering
			device.BeginScene();

			// set fill mode
			if (wireframe)
				device.RenderState.FillMode = FillMode.WireFrame;
			else
				device.RenderState.FillMode = FillMode.Solid;

			// transform camera to desired position
			device.Transform.View = rotationMatrix * Matrix.Translation(0, 0, distance);

			// set vertex format
			device.VertexFormat = MyVertex.Format;
			// render triangle mesh
			device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList, 0, vertices.Length,
				indices.Length / 3, indices, false, vertices);

			// scene is done
			device.EndScene();
			// show result
			device.Present();
		}

		/// <summary>
		/// Responds to mouse movement.
		/// </summary>
		private void RendererWindow_MouseMove(object sender, MouseEventArgs e)
		{
			// mouse control speed, lower number yields higher speed
			const float SPEED = 60;

			// determine pressed mouse button
			switch (e.Button)
			{
				case MouseButtons.Left:
					// rotate view
					float yaw = (float)(e.X - oldMouseX) / SPEED;
					float pitch = (float)(e.Y - oldMouseY) / SPEED;
					rotationMatrix *= Matrix.RotationY(yaw) * Matrix.RotationX(pitch);
					break;

				case MouseButtons.Right:
					// move camera closer / farther
					distance += (float)(e.Y - oldMouseY) / SPEED;
					break;
			}

			// update mouse cursor coordinates
			oldMouseX = e.X;
			oldMouseY = e.Y;
		}

		/// <summary>
		/// Responds to keyboard input.
		/// </summary>
		private void RendererWindow_KeyDown(object sender, KeyEventArgs e)
		{
			// determine key pressed
			switch (e.KeyCode)
			{
				case Keys.W:
					// toggle wireframe rendering
					wireframe ^= true;
					break;

				case Keys.Escape:
					// close renderer window
					this.Close();
					break;
			}
		}
	}
}