/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using System.Drawing;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using MyVertex = Microsoft.DirectX.Direct3D.CustomVertex.PositionColored;

namespace Zcu.Gsvd.DisplaceMap.Testing
{
	/// <summary>
	/// <p>Simple triangle mesh renderer that supports vertex colors.
	/// If there are no point color attributes gray color is used for each vertex.
	/// Intended mainly for testing purposes.</p>
	/// <img src="TriMeshRenderer.png" />
	/// <p>Press <i>W</i> key to toggle wireframe display.
	/// Press <i>Escape</i> to close renderer window.
	/// You can also use mouse to change view interactively.
	/// Moving the mouse with left button pressed rotates the view.
	/// Moving the mouse up and down with right button pressed to controls zoom.</p>
	/// <p><b>How it works:</b><br />
	/// Input MVE2 triangle mesh is converted into Direct3D style.
	/// An array of colored D3D vertices is created from original mesh points and
	/// triangle vertex indices are copied into an index array.
	/// Result is rendered in a separate thread to a special window.</p>
	/// </summary>
	[ModuleInfo("Ji Skla", "Simple triangle mesh renderer that supports vertex colors.\n\n"
		+ "View control:\n"
		+ "W key toggles wireframe display, Escape closes renderer window. "
		+ "Moving the mouse with left button pressed rotates the view. "
		+ "Moving the mouse up and down with right button pressed to controls zoom.",
		Assemblied = "2005-05-18", Category = "Renderer", IconName = "TriMeshRenderer.ico")]
	[PortInfo("InputMesh", "Triangle mesh to render.")]
	public class TriMeshRenderer : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Triangle mesh vertices.
		/// </summary>
		private MyVertex[] vertices;

		/// <summary>
		/// Triangle mesh vertex indices.
		/// </summary>
		private int[] indices;

		/// <summary>
		/// Point color attribute name.
		/// </summary>
		private string colorAttrName;

		/// <summary>
		/// Constructor. Creates input port.
		/// </summary>
		public TriMeshRenderer()
		{
			// set color attribute name fixed
			colorAttrName = typeof(ColorRGBA).Name;

			// add module input port
			AddInPort("InputMesh", typeof(TriangleMesh));
		}

		/// <summary>
		/// Creates a deep copy of this instance.
		/// </summary>
		/// <returns>Returns a deep copy of this instance.</returns>
		public override Zcu.Mve.Core.Module DeepCopy()
		{
			return base.DeepCopy();
		}

		/*
		/// <summary>
		/// Saves current module settings.
		/// </summary>
		/// <param name="config">XML element where to save.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			base.WriteConfig (config);
		}

		/// <summary>
		/// Loads module settings.
		/// </summary>
		/// <param name="config">XML element from where to load.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			base.ReadConfig (config);
		}

		/// <summary>
		/// Creates module setup dialog.
		/// </summary>
		/// <returns>Returns module setup dialog reference.</returns>
		public override Zcu.Mve.Core.ModuleSetup InvokeSetup()
		{
			return base.InvokeSetup ();
		}
		*/

		/// <summary>
		/// Converts Zcu.Mve.Visualization.ColorRGBA to 32-bit ARGB representation.
		/// </summary>
		/// <param name="mveColor">Color to convert.</param>
		/// <returns>Returns 32-bit ARGB representation of given color.</returns>
		private int ConvertColor(ColorRGBA mveColor)
		{
			return Color.FromArgb(
				mveColor.Alpha, mveColor.Red, mveColor.Green, mveColor.Blue
				).ToArgb();
		}

		/// <summary>
		/// Converts MVE2 triangle mesh into Direct3D triangle mesh
		/// and shows the result in an interactive 3D view.
		/// </summary>
		public override void Execute()
		{
			// get input triangle mesh
			TriangleMesh inputMesh = (TriangleMesh)GetInput("InputMesh");

			// prepare memory for D3D triangle mesh
			vertices = new MyVertex[inputMesh.Points.Count];
			indices = new int[3 * inputMesh.GetNTriangles()];

			string attribFound;
			// get point elevation attributes
			UniformDataArray pointColors = AttributeFinder.FindPointAttrib(
				inputMesh, colorAttrName, typeof(ColorRGBA), out attribFound);

			// check we found what we wanted
			// if there are no color attributes just continue
			if (pointColors != null && attribFound != colorAttrName)
			{
				// show a warning that given attribute has not been found
				Console.WriteLine(this.Name + " warning:");
				Console.WriteLine("\"" + colorAttrName + "\" point attribute not found or has wrong type.");
				Console.WriteLine("Using \"" + attribFound + "\" attribute instead.");
			}

			// proceed with mesh conversion
			ConvertMesh(inputMesh, pointColors);

			// start rendering in a separate thread
			new System.Threading.Thread(
				new System.Threading.ThreadStart(RunRendering)
				).Start();
		}

		/// <summary>
		/// Runs renderer window rendering loop.
		/// </summary>
		private void RunRendering()
		{
			// create render window
			RendererWindow renWin = new RendererWindow(vertices, indices);
			// show the window
			renWin.Show();
			// start rendering loop
			renWin.RenderingLoop();
		}

		/// <summary>
		/// Converts MVE2 triangle mesh to Direct3D triangle mesh.
		/// </summary>
		/// <param name="triMesh">MVE2 triangle mesh to convert.</param>
		/// <param name="colors">Triangle mesh point colors.</param>
		private void ConvertMesh(TriangleMesh triMesh, UniformDataArray colors)
		{
			// maximal absolute value of any coordinate
			double max = double.MinValue;

			// Pass 1 - find extreme point coordinates
			for (int i = 0; i < triMesh.Points.Count; i++)
			{
				// get current point
				Point3D p = (Point3D)triMesh.Points[i];

				// check extreme absolute value
				if (p.X > max)
					max = p.X;
				else if ( - p.X > max)
					max = - p.X;

				if (p.Y > max)
					max = p.Y;
				else if ( - p.Y > max)
					max = - p.Y;

				if (p.Z > max)
					max = p.Z;
				else if ( - p.Z > max)
					max = - p.Z;
			}

			// check max value, we don't want to divide by zero
			if (max == 0)
				max = 1;

			// check whether point colors are defined
			if (colors != null)
				// Pass 2 - copy colored points to vertices
				for (int i = 0; i < triMesh.Points.Count; i++)
				{
					// get current point
					Point3D p = (Point3D)triMesh.Points[i];

					// create colored vertex
					vertices[i] = new MyVertex(
						(float)(p.X / max), (float)(p.Y / max), (float)(p.Z / max),
						ConvertColor((ColorRGBA)colors[i]));
				}
			else
				// Pass 2 - copy points without colors to vertices
				for (int i = 0; i < triMesh.Points.Count; i++)
				{
					// get current point
					Point3D p = (Point3D)triMesh.Points[i];

					// create gray vertex
					vertices[i] = new MyVertex(
						(float)(p.X / max), (float)(p.Y / max), (float)(p.Z / max),
						Color.Gray.ToArgb());
				}

			// copy triangle vertex indices
			for (int i = 0, t = 0; t < triMesh.GetNTriangles(); i += 3, t++)
			{
				// get current triangle
				Triangle tri = (Triangle)triMesh.GetCells("Triangle")[t];

				// copy vertex indices
				indices[i] = tri.V1;
				indices[i+1] = tri.V2;
				indices[i+2] = tri.V3;
			}
		}
	}
}